# -*- coding: utf-8 -*-
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: embedsignature=True

cimport cython
cimport numpy as np
import numpy as np

from cython.parallel import prange, threadid, parallel
cimport openmp
from cythonarrays.numpy_types cimport *
from cythonarrays.array_properties import _ArrayProperties
from cythonarrays.array_shapes cimport ArrayShapes
from cythonarrays.array_shapes import ArrayShapes

from libc.math cimport exp, log


class DestinationChoiceError(ValueError):
    """
    Error in destination choice model,
    no accessible destinations found for demand
    """

class DataConsistencyError(IndexError):
    """
    Data is not consistent and could lead to IndexErrors when running the model
    with these data
    """


cdef class _WIVER(ArrayShapes):
    """
    BaseClass for WIVER model
    with n_groups and n_zones
    """

    @cython.initializedcheck(False)
    cpdef char calc_daily_trips(self) except -1:
        """
        Calc the daily trips for all groups and zones
        """
        cdef char t, r
        cdef long32 g, h
        cdef double tours, linking_trips
        self.reset_array('home_based_trips_gij')
        self.reset_array('linking_trips_gij')
        with nogil, parallel(num_threads=self.n_threads):
            t = threadid()
            # loop over calibration groups
            for g in prange(self.n_groups, schedule='guided'):
                if not self._active_g[g]:
                    continue
                with gil:
                    mode = self._mode_g[g]
                    sector = self._sector_g[g]
                    name = '{}_{}'.format(self.mode_name[mode],
                                          self.sector_short[sector])
                    self.logger.info('calculate group {} ({})'.format(g, name))
                # loop over home zones
                for h in range(self.n_zones):
                    tours = self._calc_tours(g, h)
                    if tours:
                        r = self._calc_destination_choice(t, g, h)
                        if r:
                            with gil:
                                self.raise_destination_choice_error(g, h)
                        linking_trips = self._calc_linking_trips(g, tours)
                        if linking_trips:
                            r = self._calc_linking_trip_choice(t, g, h)
                            if r:
                                with gil:
                                    self.raise_linking_trips_error(g, h)
                        self._calc_trips(t, g, h, tours, linking_trips)
                self._symmetrisize_trip_matrix(g)
        self.trips_gij[:] = (self.home_based_trips_gij +
                            self.linking_trips_gij +
                            self.return_trips_gij)
        self.trips_to_destination_gj[:] = (self.home_based_trips_gij +
                                           self.linking_trips_gij).sum(1)

    def raise_destination_choice_error(self, g, h):
        """raise a DestinationChoiceError for linking trips"""
        msg = '''No accessible destinations found for group {g} and home zone {h}'''
        raise DestinationChoiceError(msg.format(g=g, h=h))

    def raise_linking_trips_error(self, g, h):
        """raise a DestinationChoiceError for linking trips"""
        msg = '''No destinations cannot be linked for group {g} and home zone {h} because they is no accessibility between the destinations'''
        raise DestinationChoiceError(msg.format(g=g, h=h))

    def calc(self):
        """calc the daily trips and the trips by time slice"""
        self.calc_daily_trips()
        self.calc_time_series()
        self.aggregate_to_modes()
        self.calc_mean_distance()
        self.calc_mean_distance_mode()

    @cython.initializedcheck(False)
    cpdef calc_mean_distance(self):
        """calculate mean distance for groups"""
        cdef long32 g, i, j, mode, sector
        cdef double total_distance, total_trips,
        cdef double trips, home_based_trips, mean_distance
        cdef double total_distance_first_trip, total_tours,
        cdef double mean_distance_first_trip
        cdef double mean_distance_linking_trip
        cdef double total_linking_trips
        cdef double linking_trips
        cdef double total_distance_linking_trips

        for g in range(self.n_groups):
            if not self._active_g[g]:
                self._mean_distance_g[g] = self.NAN_d
                continue
            mode = self._mode_g[g]
            sector = self._sector_g[g]
            name = '{}_{}'.format(self.mode_name[mode],
                                  self.sector_short[sector])
            total_distance = 0
            total_distance_first_trip = 0
            total_distance_linking_trips = 0
            total_trips = 0
            total_linking_trips = 0
            total_tours = 0
            # loop over zones
            for i in range(self.n_zones):
                for j in range(self.n_zones):
                    trips = self._trips_gij[g, i, j]
                    home_based_trips = self._home_based_trips_gij[g, i, j]
                    linking_trips = self._linking_trips_gij[g, i, j]

                    total_distance += self._km_ij[i, j] * trips
                    total_distance_first_trip += (self._km_ij[i, j] *
                                                  home_based_trips)
                    total_distance_linking_trips += (self._km_ij[i, j] *
                                                     linking_trips)

                    total_trips += trips
                    total_tours += home_based_trips
                    total_linking_trips += linking_trips

            mean_distance = total_distance / total_trips
            mean_distance_first_trip = total_distance_first_trip / total_tours
            mean_distance_linking_trip = total_distance_linking_trips / total_linking_trips

            self.logger.info('mean distance of group {n:15} ({g:3}): '\
                '{d:5.2f} km, {t:7.0f} trips, {td:7.0f} vkm. '\
                'HomeBased: {dh:5.2f} km, {th:7.0f} trips, {tdh:7.0f} vkm. '\
                'LinkingTrips: {dl:5.2f} km, {tl:7.0f} trips, {tdl:7.0f} vkm. '.\
                format(g=g, n=name, d=mean_distance, t=total_trips,
                td=total_distance, dh=mean_distance_first_trip,
                th=total_tours, tdh=total_distance_first_trip,
                tl=total_linking_trips, tdl=total_distance_linking_trips,
                dl=mean_distance_linking_trip))
            self._mean_distance_g[g] = mean_distance


    @cython.initializedcheck(False)
    cpdef calc_mean_distance_mode(self):
        """calculate mean distance for modes"""
        cdef long32 m, i, j
        cdef double total_distance, total_trips, trips, mean_distance
        for m in range(self.n_modes):
            name = self.mode_name[m]
            total_distance = 0
            total_trips = 0
            # loop over zones
            for i in range(self.n_zones):
                for j in range(self.n_zones):
                    trips = self._trips_mij[m, i, j]
                    total_distance += self._km_ij[i, j] * trips
                    total_trips += trips
            mean_distance = total_distance / total_trips
            self.logger.info('mean distance of mode {m} ({n}): {d:0.2f}'.\
                format(m=m, n=name, d=mean_distance))
            self._mean_distance_m[m] = mean_distance

    @cython.initializedcheck(False)
    cdef double _calc_tours(self, long32 g, long32 h) nogil:
        """
        calc the number of linking trips starting in home zone h
        """
        cdef double tours
        tours = self._source_potential_gh[g, h] * self._tour_rates_g[g]
        return tours

    @cython.initializedcheck(False)
    cdef double _calc_linking_trips(self, long32 g, double tours) nogil:
        """
        calc the number of linking trips for group g
        """
        cdef double linking_trips
        linking_trips = tours * (self._stops_per_tour_g[g] - 1)
        return linking_trips

    @cython.initializedcheck(False)
    cdef double _calc_p_destination(self, long32 g, char m,
                                    long32 h, long32 j) nogil:
        """
        calc utility function for the destination choice
        """
        cdef double p
        cdef double sp = self._sink_potential_gj[g, j]
        cdef double bf = self._balancing_factor_gj[g, j]
        cdef double param = self._param_dist_g[g]
        cdef double time = self._travel_time_mij[m, h, j]
        p = sp * bf * exp(param * time)
        return p

    @cython.initializedcheck(False)
    cdef char _calc_destination_choice(self,
                                      char t,
                                      long32 g,
                                      long32 h) nogil:
        """
        calc the Destination choice

        Parameters
        ----------
        t : char
            the thread_no
        g : long32
            group no
        h : long 32
            home zone no

        Returns
        -------
        r : char
            error code, if -1 this should raise a DestinationChoiceError
        """
        cdef double p, total_weights
        cdef long32 j
        cdef char m = self._mode_g[g]

        # destination choice probability
        total_weights = 0
        for j in range(self.n_zones):
            p = self._calc_p_destination(g, m, h, j)
            self._p_destination_tij[t, h, j] = p
            total_weights += p
        if not total_weights:
            return -1
        for j in range(self.n_zones):
            self._p_destination_tij[t, h, j] /= total_weights
        return 0

    @cython.initializedcheck(False)
    cdef char _calc_linking_trip_choice(self,
                                      char t,
                                      long32 g,
                                      long32 h) nogil:
        """
        calc the Distribution for the linking trips

        Parameters
        ----------
        t : char
            the thread_no
        g : long32
            group no
        h : long 32
            home zone no

        Returns
        -------
        r : char
            error code, if -1 this should raise a DestinationChoiceError
        """
        cdef char m = self._mode_g[g]
        cdef long32 i, j
        cdef double p, pi, pj, total_weights, savings_factor

        # reset arrays
        #self._weight_links_total_t[t] = 0
        total_weights = 0
        with gil:
            self.p_links_tij[t] = 0

        # tour generation for linking trips
        # loop over destinations
        for i in range(self.n_zones):
            pi = self._p_destination_tij[t, h, i]
            if pi:
                for j in range(self.n_zones):
                    pj = self._p_destination_tij[t, h, j]
                    if pj:
                        savings_factor = self._calc_savings_factor(
                        g, m, h, i, j)
                        p = pi * pj * savings_factor
                        self._p_links_tij[t, i, j] = p
                        total_weights += p

        if not total_weights:
            return -1
        for i in range(self.n_zones):
            for j in range(self.n_zones):
                self._p_links_tij[t, i, j] /= total_weights
        return 0

    @cython.initializedcheck(False)
    cdef char _calc_trips(self,
                         char t,
                         long32 g,
                         long32 h,
                         double tours,
                         double linking_trips) nogil:

        cdef long32 i, j
        cdef double trips, p
        # Fahrten Heimatzone - 1. Stop-Heimatzone
        for i in range(self.n_zones):
            trips = tours * self._p_destination_tij[t, h, i]
            self._home_based_trips_gij[g, h, i] += trips

            if linking_trips:
                # Zwischenfahrten
                for j in range(self.n_zones):
                    p = self._p_links_tij[t, i, j]
                    if p:
                        trips = linking_trips * p
                        self._linking_trips_gij[g, i, j] += trips

    @cython.initializedcheck(False)
    cdef double _calc_savings_factor(self, long32 g, char m,
                                    long32 h, long32 i, long32 j) nogil:
        """Calc the saving factor"""

        cdef double savings, savings_factor
        cdef char s
        savings = self._calc_savings(g, m, h, i, j)
        for s in range(self.n_savings_categories):
            if savings <= self._savings_bins_s[s]:
                break
        savings_factor = self._savings_weights_gs[g, s]
        return savings_factor

    @cython.initializedcheck(False)
    cdef double _calc_savings(self, long32 g, char m,
                              long32 h, long32 i, long32 j) nogil:
        """Calc the savings"""
        cdef double savings, t_hi ,t_jh, t_ij

        t_hi = self._travel_time_mij[m, h, i]
        t_jh = self._travel_time_mij[m, j, h]
        t_ij = self._travel_time_mij[m, i, j]
        savings = (t_hi + t_jh - t_ij) / (t_hi + t_jh)
        return savings

    @cython.initializedcheck(False)
    cdef char _symmetrisize_trip_matrix(self, long32 g,) nogil:
        """
        Symmetrisize the linking trips matrix and
        Calculate the trips back home
        """
        cdef long i, j
        cdef double t1, t2, t0
        for i in range(self.n_zones):
            for j in range(self.n_zones):

                # calculate the mean of forward and backward trips
                t1 = self._linking_trips_gij[g, i, j]
                t2 = self._linking_trips_gij[g, j, i]
                t0 = (t1 + t2) / 2
                self._linking_trips_gij[g, i, j] = t0
                self._linking_trips_gij[g, j, i] = t0

                # set the return trips
                self._return_trips_gij[g, i, j] = \
                    self._home_based_trips_gij[g, j, i]


    # python wrapper functions around nogil-functions - for testing purposes
    def calc_savings(self, g, m, h, i, j):
        """
        calc the savings for group g with home zone h from zone i to j
        """
        return self._calc_savings(g, m, h, i, j)

    def calc_savings_factor(self, g, m, h, i, j):
        """
        calc the saving factor for group g with home zone h from zone i to j
        """
        return self._calc_savings_factor(g, m, h, i, j)

    def calc_tours(self, g, h):
        return self._calc_tours(g, h)

    def calc_linking_trips(self, g, tours):
        return self._calc_linking_trips(g, tours)

    def calc_p_destination(self, g, m, h, j):
        return self._calc_p_destination(g, m, h, j)

    def calc_destination_choice(self, t, g, h):
        r = self._calc_destination_choice(t, g, h)
        if r:
            self.raise_destination_choice_error(g, h)

    def calc_linking_trip_choice(self, t, g, h):
        r = self._calc_linking_trip_choice(t, g, h)
        if r:
            self.raise_linking_trips_error(g, h)

    def calc_trips(self, t, g, h, tours, linking_trips):
        return self._calc_trips(t, g, h, tours, linking_trips)

    def normalise_time_series(self, time_series):
        """normalise a time_series to ensure it adds up to 100 %"""
        time_series /= time_series.sum(-1)[:, np.newaxis]

    @cython.initializedcheck(False)
    cpdef calc_time_series(self):
        """Calc the time series"""
        self.assert_data_consistency()
        cdef long32 g, s, i, j
        cdef double trips, w_starting, w_linking, w_ending
        self.reset_array('trips_gsij')
        # normalize the weights to ensure that all time slices sum up to 100 %
        self.normalise_time_series(self.time_series_starting_trips_gs)
        self.normalise_time_series(self.time_series_linking_trips_gs)
        self.normalise_time_series(self.time_series_ending_trips_gs)

        # loop over all groups
        with nogil, parallel(num_threads=self.n_threads):
            for g in prange(self.n_groups):
                for s in range(self.n_time_slices):
                    w_starting = self._time_series_starting_trips_gs[g, s]
                    w_linking = self._time_series_linking_trips_gs[g, s]
                    w_ending = self._time_series_ending_trips_gs[g, s]
                    for i in range(self.n_zones):
                        for j in range(self.n_zones):
                            trips = (
                            self._home_based_trips_gij[g, i, j] * w_starting +
                            self._linking_trips_gij[g, i, j] * w_linking +
                            self._return_trips_gij[g, i, j] * w_ending)

                            self._trips_gsij[g, s, i, j] = trips

    @cython.initializedcheck(False)
    cpdef aggregate_to_modes(self):
        """Aggregate result matrices to mode-matrices"""
        cdef long32 g, i, j, m, s
        cdef double trips
        # reset the result arrays
        self.reset_array('trips_mij')
        self.reset_array('trips_msij')

        # loop over groups
        for g in range(self.n_groups):
            m = self._mode_g[g]
            for i in range(self.n_zones):
                for j in range(self.n_zones):
                    # add to daily matrix
                    trips = self._trips_gij[g, i, j]
                    self._trips_mij[m, i, j] += trips
                    # add to time-slice matrix
                    for s in range(self.n_time_slices):
                        trips = self._trips_gsij[g, s, i, j]
                        self._trips_msij[m, s, i, j] += trips


    def assert_data_consistency(self):
        """assert the data consistency"""
        self.assert_data_consistency_of_array('mode_g', 'n_modes')

    def assert_data_consistency_of_array(self, attrname, dim):
        """
        assert the consistency of array `attrname'

        Parameters
        ----------
        attrname : str
            the name of the attribute to test
        dim : str
            the name of the dimension to test
        """
        # modes of groups
        lbound = 0
        ubound = getattr(self, dim)
        arr = getattr(self, attrname)
        a = np.logical_or(arr < lbound, arr >= ubound)
        if a.any():
            msg = '''array {v} should only contain values v with 0 <= v < {d}={n},
but contains the following invalid values at pos {pos}:
{val}'''
            raise DataConsistencyError(msg.format(v=attrname, d=dim, n=ubound,
            pos=a.nonzero(), val=arr[a]))




from lxml import etree
import os


class XMLParamParser:
    '''
    Parses an xml parameter file (gui_vm project-style) for tdmks
    '''
    # xml-tags of nodes of interest
    SCENARIO = "Szenario"
    INPUT = "Eingabe"
    RUN = "Ergebnis"
    RESOURCE_PATH = "Projektdatei"

    # map the names of the xml tags (values) to the injectables (keys)
    MAP_INPUT = {'params_file': 'params',
                 'matrix_file': 'matrices',
                 'zones_file': 'zonal_data'
                 }

    # map the names of the xml options (values)
    MAP_FLAGS = {}
    MAP_SPEC = {}

    # default values if params are not set in xml
    DEFAULTS = {}

    def __init__(self, filename, scenario_name, run_name):
        tree = etree.parse(filename)
        self._filename = filename
        self._root = tree.getroot()
        self._scenario = self._get_scenario(self._root, scenario_name)
        if self._scenario is None:
            raise Exception('Kein Szenario mit dem Namen {} gefunden!'
                            .format(scenario_name))
        self._dict = self._build_dict(run_name)

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def _build_dict(self, run_name):
        inputs = self._get_inputs(self._scenario)
        run = self._get_run(self._scenario, run_name)
        if run is None:
            raise Exception('Kein Lauf mit dem Namen {} im Szenario {} gefunden!'
                            .format(run_name, self._scenario.attrib['name']))
        d = {}

        #Input files
        for key, value in self.MAP_INPUT.items():
            path = self._get_path(inputs, value)
            if path is None:
                path = self.DEFAULTS[key]
            d[key] = path

        #Flags (like calibrate)
        for key, value in self.MAP_FLAGS.items():
            flag = self._get_option(run, value)
            if flag is None:
                d[key] = self.DEFAULTS[key]
            elif flag and flag.strip().lower() == 'true':
                d[key] = True
            else:
                d[key] = False

        #Options (like groups)
        for key, value in self.MAP_SPEC.items():
            option = self._get_option(run, value)
            if option is not None:
                option = option.strip()
            else:
                option = self.DEFAULTS[key]
            d[key] = option

        # results file
        path = self._get_path([run], run_name)
        if path is None:
            path = scenario_name + '.h5'
        d['result_file'] = path

        return d

    def _get_scenario(self, element, scenario_name):
        '''
        get scenario by name, None if not found

        Params
        ------
        scenario_name - String, name of the scenario

        Return
        ------
        inputs - list of etree.Element, all inputs of given scenario
        '''
        for scenario in element.findall('.//' + self.SCENARIO):
            scen_name = scenario.attrib['name']
            if scen_name == scenario_name:
                return scenario
        return None

    def _get_inputs(self, scenario):
        '''
        get all inputs of given scenario node

        Params
        ------
        scenario - etree.Element, describes scenario and it's configuration

        Return
        ------
        inputs - list of etree.Element, all inputs of given scenario
        '''
        return scenario.findall( './/' + self.INPUT)

    def _get_run(self, scenario, run_name):
        '''
        get a run of the given scenario with the given name

        Params
        ------
        scenario - etree.Element, describes scenario and it's configuration

        Return
        ------
        run - etree.Element, the run
        '''
        for run in scenario.findall('.//' + self.RUN):
            r_name = run.attrib['name']
            if r_name == run_name:
                return run
        return None

    def _get_path(self, resources, resource_name):
        '''
        gets filepath of given resource (input or ), None if not found

        Params
        ------
        resource_name - String, name of the resource

        Return
        ------
        file_path - String, path to the resource file
        '''
        for res in resources:
            if res.attrib['name'] == resource_name:
                path = res.find(self.RESOURCE_PATH)
                if path.attrib.has_key('relative') and path.attrib['relative']:
                    return self._get_absolute_path(res)
                return path.text
        return None

    def _get_option(self, run, option_name):
        options = run.findall('Option')
        for opt in options:
            if opt.attrib['name'] == option_name:
                return opt.text
        return None

    def _get_absolute_path(self, resource):
        typ = resource.tag
        rel_path = resource.find(self.RESOURCE_PATH).text
        project_path = os.path.split(self._filename)[0]
        scen_name = self._scenario.attrib['name']
        #ToDo: map this
        if typ == self.INPUT:
            sub = 'Eingaben'
        elif typ == self.RUN:
            sub = 'Ausgaben'
        else:
            sub = ''
        full_path = os.path.join(project_path, scen_name, sub, rel_path)
        return full_path


# -*- coding: utf-8 -*-

import multiprocessing
import os
from collections import defaultdict
import numpy as np
import pandas as pd
import xarray as xr


from cythonarrays.array_properties import _ArrayProperties
from matrixconverters.save_ptv import SavePTV
from matrixconverters.xarray2netcdf import xr2netcdf
import pyximport; pyximport.install()
from wiver.wiver_cython import (_WIVER,
                                DestinationChoiceError,
                                DataConsistencyError)

np.seterr(divide='ignore', invalid='ignore')


class WIVER(_WIVER, _ArrayProperties):
    """WIVER Commercial Trips Model"""
    _coordinates = {'n_groups': 'groups',
                    'n_modes': 'modes',
                    'n_zones': 'zone_name',
                    'n_time_slices': 'lbl_time_slice',
                    'n_sectors': 'sector_short',
                    }

    def __init__(self,
                 n_groups: int,
                 n_zones: int,
                 n_savings_categories: int=9,
                 n_time_slices: int=5,
                 n_modes: int=4,
                 n_sectors: int=2,
                 n_threads: int=None):
        super().__init__()

        self.n_modes = n_modes
        self.n_groups = n_groups
        self.n_zones = n_zones
        self.n_sectors = n_sectors
        self.n_savings_categories = n_savings_categories
        self.n_time_slices = n_time_slices
        self.set_n_threads(n_threads=n_threads)

        self.define_arrays()
        self.init_arrays()

    def set_n_threads(self, n_threads: int=None):
        """
        Set the number of threads

        Parameters
        ----------
        n_threads: int, optional
        """
        if n_threads:
            n_threads = min(n_threads, multiprocessing.cpu_count())
        else:
            n_threads = multiprocessing.cpu_count()
        self.n_threads = min(n_threads, self.n_groups)

    @classmethod
    def read_from_netcdf(cls, files, n_threads=None):
        """Read a Wiver Model
        from a set of netcdf-Filename located in folder"""
        # create instance of self
        self = cls(n_groups=0, n_zones=0, n_threads=n_threads)
        # add datasets
        self.read_all_data(files)
        self.data = xr.merge((self.params, self.matrices, self.zonal_data,
                              self.balancing))

        # set the dimensions
        dims = self.data.dims
        self.n_zones = dims['origins']
        self.n_groups = dims['groups']
        self.n_modes = dims['modes']
        self.n_savings_categories = dims['savings']
        self.n_time_slices = dims['time_slices']
        self.set_n_threads()

        # resize the arrays to the right dimensions
        self.init_arrays()

        self.set_arrays_from_dataset()

        return self

    def set_arrays_from_dataset(self):
        """Sets the arrays with values from the dataset"""
        ds = self.data

        # params
        self.modes = ds.modes.data
        self.groups = ds.groups.data
        self.mode_g = ds.mode_of_groups.data

        self.n_sectors = len(ds.sector_short)

        self.sector_g = ds.sector_of_groups.data
        self.mode_name = ds.mode_name.data
        self.sector_short = ds.sector_short.data
        self.param_dist_g = ds.param_dist.data
        self.savings_bins_s = ds.savings_bins.data
        self.savings_weights_gs = ds.savings_weights.data
        self.tour_rates_g = ds.tour_rates.data
        self.stops_per_tour_g = ds.stops_per_tour.data
        self.time_series_starting_trips_gs = ds.time_series_starting_trips.data
        self.time_series_linking_trips_gs = ds.time_series_linking_trips.data
        self.time_series_ending_trips_gs = ds.time_series_ending_trips.data

        # zonal data
        self.source_potential_gh = ds.source_potential.data
        self.sink_potential_gj = ds.sink_potential.data
        self.zone_no = ds.zone_no.data
        self.zone_name = ds.zone_name.data

        # matrices
        self.travel_time_mij = ds.travel_time.data
        self.km_ij = ds.distance_matrix.data

        self.set_arrays_from_balancing_ds(ds)

        # results
        self.results = self.define_results()

    def set_arrays_from_balancing_ds(self, ds):
        """Set arrays from balancing dataset"""
        # balancing factors
        self.balancing_factor_gj = ds.balancing_factor.data
        self.trips_to_destination_gj = ds.trips_to_destination.data

    def define_arrays(self):
        """Define the arrays"""
        self.init_object_array('groups', 'n_groups')
        self.init_object_array('group_names', 'n_groups')
        self.init_object_array('modes', 'n_modes')
        self.init_object_array('mode_name', 'n_modes')
        self.init_object_array('sector_short', 'n_sectors')
        self.init_object_array('zone_name', 'n_zones')
        self.init_object_array('lbl_time_slice', 'n_time_slices')

        self.init_array('mode_g', 'n_groups', 0)
        self.init_array('sector_g', 'n_groups', 0)
        self.init_array('active_g', 'n_groups', 1)

        self.init_array('savings_bins_s', 'n_savings_categories')
        self.init_array('savings_weights_gs',
                        'n_groups, n_savings_categories', 1)

        self.init_array('km_ij', 'n_zones, n_zones')
        self.init_array('travel_time_mij', 'n_modes, n_zones, n_zones')
        self.init_array('mean_distance_g', 'n_groups')
        self.init_array('mean_distance_m', 'n_modes')
        self.init_array('param_dist_g', 'n_groups', -0.1)
        self.init_array('tour_rates_g', 'n_groups', 1)
        self.init_array('stops_per_tour_g', 'n_groups', 2)

        self.init_array('source_potential_gh', 'n_groups, n_zones')
        self.init_array('sink_potential_gj', 'n_groups, n_zones')
        self.init_array('balancing_factor_gj', 'n_groups, n_zones', 1)
        self.init_array('trips_to_destination_gj', 'n_groups, n_zones', 0)

        self.init_array('trips_gij', 'n_groups, n_zones, n_zones', 0)
        self.init_array('home_based_trips_gij',
                        'n_groups, n_zones, n_zones', 0)
        self.init_array('linking_trips_gij', 'n_groups, n_zones, n_zones', 0)
        self.init_array('return_trips_gij', 'n_groups, n_zones, n_zones', 0)
        self.init_array('p_destination_tij', 'n_threads, n_zones, n_zones', 0)
        self.init_array('p_links_tij', 'n_threads, n_zones, n_zones', 0)

        self.init_array('trips_gsij',
                        'n_groups, n_time_slices, n_zones, n_zones', 0)

        self.init_array('trips_mij',
                        'n_modes, n_zones, n_zones', 0)
        self.init_array('trips_msij',
                        'n_modes, n_time_slices, n_zones, n_zones', 0)

        self.init_array('time_series_starting_trips_gs',
                        'n_groups, n_time_slices', 1)
        self.init_array('time_series_linking_trips_gs',
                        'n_groups, n_time_slices', 1)
        self.init_array('time_series_ending_trips_gs',
                        'n_groups, n_time_slices', 1)

    def define_datasets(self):
        """Define the datasets"""
        self.params = self.define_params()
        self.matrices = self.define_matrices()
        self.zonal_data = self.define_zonal_data()
        self.balancing = self.define_balancing()
        self.results = self.define_results()

    def define_params(self):
        """Define the params"""
        ds = xr.Dataset()
        ds['modes'] = self.modes
        ds['mode_name'] = (('modes'), self.mode_name)
        ds['groups'] = self.groups
        ds['mode_of_groups'] = (('groups'),
                                self.mode_g)
        ds['sector_of_groups'] = (('groups'),
                                    self.sector_g)
        ds['sector_short'] = (('sectors'), self.sector_short)

        # assign group names
        for g in range(self.n_groups):
            self.group_names[g] = '{}_{}'.format(
                self.mode_name[self.mode_g[g]],
                self.sector_short[self.sector_g[g]])

        ds['group_names'] = (('groups'), self.group_names)

        ds['param_dist'] = (('groups'),
                            self.param_dist_g)
        ds['savings_bins'] = (('savings'),
                              self.savings_bins_s)
        ds['savings_weights'] = (('groups', 'savings'),
                                 self.savings_weights_gs)
        ds['tour_rates'] = (('groups'),
                            self.tour_rates_g)
        ds['stops_per_tour'] = (('groups'),
                                self.stops_per_tour_g)
        ds['time_series_starting_trips'] = (
            ('groups', 'time_slices'), self.time_series_starting_trips_gs)
        ds['time_series_linking_trips'] = (
            ('groups', 'time_slices'), self.time_series_linking_trips_gs)
        ds['time_series_ending_trips'] = (
            ('groups', 'time_slices'), self.time_series_ending_trips_gs)
        return ds

    def define_balancing(self):
        """Define the balancing factors"""
        ds = xr.Dataset()
        ds['balancing_factor'] = (('groups', 'destinations'),
                                    self.balancing_factor_gj)
        ds['trips_to_destination'] = (('groups', 'destinations'),
                                      self.trips_to_destination_gj)
        return ds

    def define_zonal_data(self):
        """Define the matrices"""
        ds = xr.Dataset()
        ds['groups'] = self.groups
        ds['zone_no'] = self.zone_no
        ds['zone_name'] = (('zone_no'), self.zone_name)
        ds['origins'] = self.zone_no
        ds['destinations'] = self.zone_no
        ds['source_potential'] = (('groups', 'origins'),
                                  self.source_potential_gh)
        ds['sink_potential'] = (('groups', 'destinations'),
                                self.sink_potential_gj)
        return ds

    def define_matrices(self):
        """Define the matrices"""
        ds = xr.Dataset()
        ds['modes'] = self.modes
        ds['origins'] = self.zone_no
        ds['destinations'] = self.zone_no
        ds['zone_no'] = self.zone_no
        ds['zone_name'] = (('zone_no'), self.zone_name)
        ds['travel_time'] = (('modes', 'origins', 'destinations'),
                             self.travel_time_mij)
        ds['distance_matrix'] = (('origins', 'destinations'),
                                 self.km_ij)

        return ds

    def define_results(self):
        """Define the results"""
        ds = xr.Dataset()
        # resulting arrays
        ds['modes'] = self.modes
        ds['groups'] = self.groups
        ds['origins'] = self.zone_no
        ds['destinations'] = self.zone_no
        ds['zone_no'] = self.zone_no
        ds['zone_name'] = (('zone_no'), self.zone_name)
        ds['trips_gij'] = (('groups', 'origins', 'destinations'),
                           self.trips_gij)
        ds['trips_gsij'] = (('groups', 'time_slices',
                             'origins', 'destinations'),
                            self.trips_gsij)
        ds['trips_mij'] = (('modes', 'origins', 'destinations'),
                           self.trips_mij)
        ds['trips_msij'] = (('modes', 'time_slices',
                             'origins', 'destinations'),
                            self.trips_msij)
        ds['mean_distance_groups'] = (('groups',),
                                      self.mean_distance_g)
        ds['mean_distance_modes'] = (('modes',),
                                     self.mean_distance_m)

        return ds

    def merge_datasets(self):
        """Merge the datasets"""
        self.data = xr.merge((self.params, self.zonal_data,
                              self.matrices, self.balancing,
                              self.results))

    def save_all_data(self, wiver_files):
        """Save Dataset to netcdf-file"""
        datasets = ('params', 'zonal_data', 'matrices',
                    'balancing', 'results')
        for dataset_name in datasets:
            fn = wiver_files[dataset_name]
            self.save_data(dataset_name, fn)

    def save_data(self, dataset_name, fn):
        """Save Dataset to netcdf-file"""
        ds = getattr(self, dataset_name)
        self.logger.info('write {}'.format(fn))
        dirname = os.path.dirname(fn)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        xr2netcdf(ds, fn)
        ds.close()

    def read_all_data(self, datasets):
        """Read Datasets from netcdf-file"""
        for dataset_name in ('params', 'matrices', 'zonal_data' ,
                             'balancing'):
            fn = datasets[dataset_name]
            self.read_data(dataset_name, fn)

    def read_data(self, dataset_name, fn):
        """read single dataset from """
        self.logger.info('read {}'.format(fn))
        ds = xr.open_dataset(fn).load()
        ds.close()
        setattr(self, dataset_name, ds)

    def save_results(self, wiver_files):
        """Save results except trips_gsij to folder"""
        del self.results['trips_gsij']
        dataset_name = 'results'
        fn = wiver_files[dataset_name]
        self.save_data(dataset_name, fn)
        dataset_name = 'balancing'
        fn = wiver_files[dataset_name]
        self.save_data(dataset_name, fn)

    def save_results_to_visum(self, folder, visum_format='BK'):
        """Save the results to VISUM-Format"""
        for m, mode in enumerate(self.modes):
            visum_ds = xr.Dataset()
            visum_ds['matrix'] = self.results.trips_mij[m]
            visum_ds['zone_no'] = self.zone_no
            visum_ds['zone_name'] = self.zone_name
            #visum_ds['zone_names2'] = self.zone_name
            s = SavePTV(visum_ds)
            file_name = os.path.join(
                folder, '{m}.mtx'.format(m=mode))
            self.logger.info('save matrix for mode {m} to {f}'.format(
                m=mode, f=file_name
            ))
            s.savePTVMatrix(file_name, Ftype=visum_format)

    def save_detailed_results_to_visum(self, folder, visum_format='BK'):
        """Save the results to VISUM-Format"""
        sectors = defaultdict(list)
        for g, group in enumerate(self.groups):
            if self.active_g[g]:
                sector_id = group % 100
                sectors[sector_id].append(g)
        self.logger.info('sectors: {}'.format(sectors))
        for sector_id, sector_groups in sectors.items():
            self.logger.info('sector_id: {}, groups: {}'.format(sector_id,
                                                                sector_groups))
            name = self.params.sector_short.sel(sectors=sector_id).values
            self.logger.info('name: {}'.format(name))
            visum_ds = xr.Dataset()
            visum_ds['zone_no'] = self.zone_no
            visum_ds['zone_names'] = self.zone_name
            matrix = 0
            self.logger.info('Sector {s}: add wiver-groups'.format(s=sector_id))
            for g in sector_groups:
                mat = self.results.trips_gij[g]
                mode = self.mode_g[g]
                mode_descr = self.modes[mode]
                group_name = self.group_names[g]
                self.logger.info(
                    'add group {g} {m}: {s:.0f}'.format(g=group_name,
                                                        m=mode_descr,
                                                        s=float(mat.sum())))
                matrix += mat
            visum_ds['matrix'] = matrix
            self.logger.info('Sector {s}: {t:.0f} trips'.format(
                s=sector_id, t=float(matrix.sum())
            ))
            s = SavePTV(visum_ds)
            file_name = os.path.join(
                folder, 'wiver_{sector_id}_{n}.mtx'.format(
                    sector_id=sector_id, n=name))
            self.logger.info('save matrix for sector {s}_{n} to {f}'.format(
                s=sector_id, n=name, f=file_name
            ))
            s.savePTVMatrix(file_name, Ftype=visum_format)
            self.logger.info('matrix_saved')

    def adjust_balancing_factor(self, threshold=0.1):
        """
        Randsummenabgleich für Zielpotenziale
        """
        self.converged=False
        sp = self.zonal_data.sink_potential
        target_share = sp / sp.sum('destinations')
        trips = self.balancing.trips_to_destination
        actual_share = trips / trips.sum('destinations')
        kf = target_share / actual_share
        kf[:] = kf.fillna(1)
        bf = self.balancing.balancing_factor
        # adjust balancing factor
        bf[:] = bf.fillna(1)
        bf[:] = bf * kf
        # normalize balancing factor
        #bf[:] = bf / bf.mean('destinations')
        if (np.abs(kf - 1) < threshold).all():
            self.converged = True
            self.logger.info('converged!')

    def calc_with_balancing(self, max_iterations=10, threshold=0.1):
        """calculate with balancing the """
        self.converged = False
        iteration = 0
        while (not self.converged) and iteration < max_iterations:
            iteration += 1
            self.logger.info('calculate trips in iteration {}'.format(iteration))
            self.calc()
            self.logger.info('Total trips: {:0.2f}'.format(self.trips_gij.sum()))
            self.adjust_balancing_factor(threshold)

    def calc_starting_and_ending_trips(self):
        """
        calculate the starting trips per zone and group
        """
        trips_ij = self.results.trips_gij.sum('groups')
        modelled_trips = (trips_ij.sum('origins').\
            rename({'destinations': 'zone_no',}) + \
            trips_ij.sum('destinations').rename({'origins': 'zone_no',})) / 2

        df_modelled = modelled_trips.to_dataframe(name='modelled_trips')

        group_labels = []
        for g, group in enumerate(self.groups):
            sn = self.data.sector_short[int(self.data.sector_of_groups[g])].data
            mn = self.data.mode_name[int(self.data.mode_of_groups[g])].data
            group_label = '{}_{}'.format(mn, sn).replace(' ', '')
            group_labels.append(group_label)

        sp = self.data.sink_potential

        starting_trips_gh = (self.data.source_potential * self.data.tour_rates)
        if 'zone_no' in starting_trips_gh.coords:
            starting_trips_gh = starting_trips_gh.drop(['zone_no'])
        starting_trips_gh = starting_trips_gh.rename(
            {'origins': 'zone_no',})

        ending_trips_g = starting_trips_gh.sum('zone_no') * self.data.stops_per_tour
        ending_trips_gh = sp * \
            (ending_trips_g / sp.sum('destinations'))

        starting_trips_i = starting_trips_gh.sum('groups')
        ending_trips_i = ending_trips_gh.sum('groups')

        df_balancing = self.data.balancing_factor.to_dataframe(name='balance')
        df_balancing = df_balancing.reset_index().pivot(
            index='destinations', columns='groups', values='balance')
        df_balancing.columns =['Bal_{}'.format(col)
                                  for col in group_labels]

        df_s = starting_trips_i.to_dataframe(name='starting_trips')
        df_e = ending_trips_i.to_dataframe(name='ending_trips')
        df_e.index.name = 'zone_no'
        df_modelled['start_end'] = df_s.starting_trips + df_e.ending_trips

        df_ending = ending_trips_gh.to_dataframe(name='ending_trips_g')
        df_ending = df_ending.reset_index().pivot(
            index='destinations', columns='groups', values='ending_trips_g')

        df_ending.index.name = 'zone_no'
        df_ending.columns =['End_{}'.format(col)
                                for col in group_labels]

        df_starting = starting_trips_gh.to_dataframe(name='starting_trips_g')
        df_starting = df_starting.reset_index().pivot(
            index='zone_no', columns='groups', values='starting_trips_g')
        df_starting.columns =['Start_{}'.format(col)
                                for col in group_labels]

        df_sp = self.data.sink_potential.to_dataframe(name='sink_potential')
        df_sp = df_sp.reset_index().\
            rename(columns={'destinations': 'zone_no',}).\
            pivot(index='zone_no', columns='groups', values='sink_potential')
        df_sp.columns =['SP_{}'.format(col)
                        for col in group_labels]

        df_name = self.data.zone_name.to_dataframe()

        df = pd.concat([df_name,
                        df_modelled,
                        df_s,
                        df_e,
                        df_starting,
                        df_ending,
                        df_sp,
                        df_balancing],
                       axis=1)
        return df




# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import os
import orca
import tempfile
from argparse import ArgumentParser
from wiver.wiver_python import (WIVER)
from wiver.wiver_xml_param_parser import XMLParamParser
from wiver.wiver import (project_folder,
                         params_file,
                         matrix_file,
                         zones_file,
                         result_file,
                         wiver,
                         run_wiver,
                         save_results,
                         )


def main():
    parser = ArgumentParser(description="Travel Demand Model")

    parser.add_argument("-xml", action="store",
                        help="XML parameter file",
                        dest="xml_params", default=None)

    parser.add_argument("-n", action="store",
                        help="Scenario Name",
                        dest="scenario_name", default='BaseScenario')

    parser.add_argument("-r", action="store",
                        help="name of the run",
                        dest="run_name", default='Gesamtlauf')

    options = parser.parse_args()

    x = XMLParamParser(options.xml_params,
                       options.scenario_name,
                       options.run_name)

    for key in x:
        value = x[key]
        if value is not None:
            orca.add_injectable(key, value)

    orca.run(['run_wiver', 'save_results'])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from cythonarrays.numpy_types cimport *

from cythonarrays.array_shapes cimport ArrayShapes
from cythonarrays.array_shapes import ArrayShapes


cdef class _WIVER(ArrayShapes):
    """
    BaseClass for WIVER model
    with n_groups and n_zones
    """
    cdef public long32 n_groups
    cdef public long32 n_zones
    cdef public char n_threads
    cdef public char n_savings_categories
    cdef public char n_time_slices
    cdef public char n_modes
    cdef public char n_sectors

    # mode and sector of groups
    cdef public ARRAY_1D_i1 _mode_g
    cdef public ARRAY_1D_i1 _sector_g

    # active groups
    cdef public ARRAY_1D_i1 _active_g

    # Savings bins and weights for each group
    cdef public ARRAY_1D_d _savings_bins_s
    cdef public ARRAY_2D_d _savings_weights_gs

    # distance matrix (km from a to b)
    cdef public ARRAY_2D_d _km_ij

    # skim matrix (travel time from a to b)
    cdef public ARRAY_3D_d _travel_time_mij

    # Distance Parameter for each group
    cdef public ARRAY_1D_d _param_dist_g

    # Mean Distance for each group
    cdef public ARRAY_1D_d _mean_distance_g
    # Mean Distance for each mode
    cdef public ARRAY_1D_d _mean_distance_m

    # Tour rates and number of stops per tour by group
    cdef public ARRAY_1D_d _tour_rates_g
    cdef public ARRAY_1D_d _stops_per_tour_g

    # Sources and Sinks for each group
    cdef public ARRAY_2D_d _source_potential_gh
    cdef public ARRAY_2D_d _sink_potential_gj
    cdef public ARRAY_2D_d _balancing_factor_gj

    # time series for each group
    cdef public ARRAY_2D_d _time_series_starting_trips_gs
    cdef public ARRAY_2D_d _time_series_linking_trips_gs
    cdef public ARRAY_2D_d _time_series_ending_trips_gs

    # resulting trips
    cdef public ARRAY_3D_d _trips_gij
    cdef public ARRAY_3D_d _home_based_trips_gij
    cdef public ARRAY_3D_d _linking_trips_gij
    cdef public ARRAY_3D_d _return_trips_gij

    cdef public ARRAY_2D_d _trips_to_destination_gj

    # Trips by time slice
    cdef public ARRAY_4D_d _trips_gsij

    # trips_by_mode
    cdef public ARRAY_3D_d _trips_mij
    cdef public ARRAY_4D_d _trips_msij

    # internal probability matrices for each thread
    cdef public ARRAY_3D_d _p_destination_tij
    cdef public ARRAY_3D_d _p_links_tij

    cpdef char calc_daily_trips(self) except -1
    cdef double _calc_tours(self, long32 g, long32 h) nogil
    cdef double _calc_linking_trips(self, long32 g, double tours) nogil
    cdef double _calc_p_destination(self, long32 g, char m,
                                    long32 h, long32 j) nogil
    cdef char _calc_destination_choice(self,
                                      char t,
                                      long32 g,
                                      long32 h) nogil

    cdef char _calc_linking_trip_choice(self,
                                      char t,
                                      long32 g,
                                      long32 h) nogil

    cdef char _symmetrisize_trip_matrix(self, long32 g,) nogil

    cdef char _calc_trips(self,
                         char t,
                         long32 g,
                         long32 h,
                         double tours,
                         double linking_trips) nogil
    cdef double _calc_savings_factor(self, long32 g, char m,
                                   long32 h, long32 i, long32 j) nogil
    cdef double _calc_savings(self, long32 g, char m,
                             long32 h, long32 i, long32 j) nogil

    cpdef calc_time_series(self)
    cpdef aggregate_to_modes(self)
    cpdef calc_mean_distance(self)
    cpdef calc_mean_distance_mode(self)


# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import tempfile
import os
from argparse import ArgumentParser
from typing import List, Dict
import sqlite3 as db
import numpy as np
import pandas as pd
from openpyxl import load_workbook

import orca
from wiver.wiver_python import WIVER
from cythonarrays.configure_logger import SimLogger


@orca.injectable()
def project_folder() -> str:
    """ The Project folder (%TEMP% by default)

    Returns
    -------
    project_folder : str
    """
    folder = tempfile.gettempdir()
    return folder


@orca.injectable()
def params_file(project_folder:str) -> str:
    """ The params-file

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    params_file : str
    """
    fn = 'params'
    file_path = os.path.join(project_folder, '{}.h5'.format(fn))
    return file_path


@orca.injectable()
def result_folder(project_folder:str) -> str:
    """The result folder

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    result_folder : str
"""
    dirname = 'matrices'
    file_path = os.path.join(project_folder, dirname)
    return file_path


@orca.injectable()
def matrix_file(project_folder:str) -> str:
    """ The matrix-file

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    matrix_file : str
    """
    fn = 'matrices'
    file_path = os.path.join(project_folder, '{}.h5'.format(fn))
    return file_path


@orca.injectable()
def zones_file(project_folder:str) -> str:
    """ The zonal_data-file

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    zones_file : str
    """
    fn = 'zonal_data'
    file_path = os.path.join(project_folder, '{}.h5'.format(fn))
    return file_path


@orca.injectable()
def result_file(project_folder:str) -> str:
    """ The results-file

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    result_file : str
    """
    fn = 'results'
    file_path = os.path.join(project_folder, '{}.h5'.format(fn))
    return file_path


@orca.injectable()
def balancing_file(project_folder:str) -> str:
    """ The balancing-file

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    balancing_file : str
    """
    fn = 'balancing'
    file_path = os.path.join(project_folder, '{}.h5'.format(fn))
    return file_path


@orca.injectable()
def starting_ending_trips_file(project_folder:str) -> str:
    """ The Excel-File for starting and ending trips

    Parameters
    ----------
    project_folder : str

    Returns
    -------
    starting_ending_trip_file : str
    """
    fn = 'starting_ending_trips.xlsx'
    file_path = os.path.join(project_folder, fn)
    return file_path


@orca.injectable()
def max_iterations() -> str:
    """ maximum number of iterations"""
    return 1


@orca.injectable()
def wiver_files(params_file:str,
                matrix_file:str,
                zones_file:str,
                balancing_file:str,
                result_file:str) -> Dict[str, str]:
    """
    Returns
    -------
    wiver_files : dict
        a dictionary with the file path to the input data files
    """
    files = {'params': params_file,
             'matrices': matrix_file,
             'zonal_data': zones_file,
             'balancing': balancing_file,
             'results': result_file}
    return files


@orca.injectable()
def n_threads() -> int:
    """
    The number of threads to use
    if not defined, use the maximum number of threads available,
    maximum number equals to the number of groups
    """


@orca.injectable()
def wiver(wiver_files: Dict[str, str], n_threads: int) -> WIVER:
    """

    Parameters
    ----------
    wiver_files : dict

    Returns
    -------
    wiver : Wiver-instace
    """
    wiver = WIVER.read_from_netcdf(wiver_files, n_threads=n_threads)
    return wiver


@orca.injectable()
def scenario() -> str:
    """The Scenario Name"""
    return 'wiver'


@orca.injectable()
def groups_to_calculate() -> List[int]:
    """List of group ids to calculate"""
    return []


@orca.injectable()
def connection(project_folder: str) -> db.Connection:
    """database connection to write zonal data into"""
    fn = os.path.join(project_folder, 'wiver.db3')
    connection = db.connect(fn)
    return connection


@orca.injectable()
def reset_balancing() -> bool:
    """if True, balancing factors will be reset to 1 before calculating"""
    return False


@orca.step()
def add_logfile(project_folder: str, scenario: str):
    """
    add Logfile to logger

    Parameters
    ----------
    wiver model
    """
    logger = SimLogger()
    logger.add_packages(['wiver'])
    logfile = os.path.join(project_folder, 'log')
    logger.configure(logfile, scenario=scenario)


@orca.step()
def run_wiver(wiver: WIVER, wiver_files: dict, max_iterations: int):
    """
    calculate wiver model

    Parameters
    ----------
    wiver: wiver-model
    wiver-files : dict
    max_iterations: int
    """
    wiver.calc_with_balancing(max_iterations=max_iterations)
    wiver.save_results(wiver_files)


@orca.step()
def run_wiver_for_selected_groups(wiver: WIVER,
                                  wiver_files: dict,
                                  max_iterations: int,
                                  reset_balancing: bool,
                                  groups_to_calculate: List[int]):
    """
    calculate wiver model for selected groups only

    Parameters
    ----------
    wiver: wiver-model
    wiver-files : dict
    max_iterations: int
    groups_to_calculate: list of int
    """
    if reset_balancing:
        wiver.init_array('balancing_factor_gj')
        wiver.init_array('trips_to_destination_gj')

    if groups_to_calculate:
        wiver.active_g[:] = np.in1d(wiver.groups, groups_to_calculate)
    wiver.calc_with_balancing(max_iterations=max_iterations)
    wiver.save_results(wiver_files)

@orca.step()
def save_results(wiver: WIVER, wiver_files: dict, result_folder: str):
    """
    save result matrices of wiver-model

    Parameters
    ----------
    wiver: wiver-model
    wiver-files : dict
    result_folder: str
        the folder to store the calculated matrices
    """
    fn = wiver_files['results']
    wiver.read_data('results', fn)
    wiver.save_results_to_visum(result_folder, visum_format='BK')


@orca.step()
def save_detailed_results(wiver: WIVER,
                          wiver_files: dict,
                          result_folder: str,
                          groups_to_calculate: List[int]):
    """
    save detailed result matrices of wiver-model

    Parameters
    ----------
    wiver: wiver-model
    wiver-files : dict
    result_folder: str
        the folder to store the calculated matrices
    groups_to_calculate: list of int
    """
    if groups_to_calculate:
        wiver.active_g[:] = np.in1d(wiver.groups, groups_to_calculate)
    fn = wiver_files['results']
    wiver.read_data('results', fn)
    wiver.save_detailed_results_to_visum(result_folder, visum_format='BK')

@orca.step()
def calc_starting_ending_trips(wiver: WIVER,
                               wiver_files: dict,
                               starting_ending_trips_file: str,
                               connection: db.Connection):
    """
    calculate starting and ending trips per zone

    Parameters
    ----------
    wiver: wiver-model
    wiver-files : dict
    starting_ending_trips_file: str
    """
    fn = wiver_files['results']
    wiver.read_data('results', fn)
    df = wiver.calc_starting_and_ending_trips()
    df.to_sql(name='wiver', con=connection, if_exists='replace')

    sheetname = 'data'
    # write to existing excel-file
    if os.path.isfile(starting_ending_trips_file):
        book = load_workbook(starting_ending_trips_file)
        with pd.ExcelWriter(starting_ending_trips_file, engine='openpyxl') as writer:
            writer.book = book
            #  overwrite the sheet if exists
            if sheetname in book.sheetnames:
                writer.book.remove(writer.book.get_sheet_by_name(sheetname))
            df.to_excel(writer, sheet_name=sheetname)
    # otherwise create a new xlsx-file
    else:
        df.to_excel(starting_ending_trips_file, sheet_name=sheetname)


if __name__ == '__main__':
    parser = ArgumentParser(description="Commercial Trip Model Wiver")
    parser.add_argument('-f', '--folder', dest='project_folder',
                        help='Project folder',
                        required=True)
    parser.add_argument('-m', '--result-folder', dest='result_folder',
                        help='Result Matrix folder',
                        required=True)
    parser.add_argument('-s', '--scenario', dest='scenario',
                        help='Scenario Name', default='Wiver',)
    parser.add_argument('-i', '--iterations', dest='max_iterations',
                        help='Maximum number of iterations', type=int,
                        default=5,)
    parser.add_argument('-g', '--groups', dest='groups_to_calculate',
                        help='groups to calculate', type=int, nargs='+')
    parser.add_argument('-c', '--calc-starting', dest='calc_starting',
                        action='store_true',
                        help='calculate starting and ending trips', )
    parser.add_argument('-d', '--save-detailed', dest='save_detailed',
                        action='store_true',
                        help='save detailed results', )
    parser.add_argument('-r', '--reset-balancing', dest='reset_balancing',
                        action='store_true',
                        help='reset balancing parameters to 0', )

    options = parser.parse_args()
    for key, value in options._get_kwargs():
        orca.add_injectable(key, value)

    steps = [
        'add_logfile',
        ]

    if options.groups_to_calculate:
        steps.append('run_wiver_for_selected_groups')
    else:
        steps.append('run_wiver')

    steps.append('save_results')

    if options.save_detailed:
        steps.append('save_detailed_results')

    if options.calc_starting:
        steps.append('calc_starting_ending_trips')

    orca.run(steps)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import os
import sys
import runpy
import shutil
import tempfile
import pytest

import xarray as xr
import numpy as np
import pandas as pd
import orca
from wiver.wiver_python import (WIVER,
                                DestinationChoiceError, DataConsistencyError)


@pytest.fixture(scope='class', params=range(2))
def group(request):
    """
    different values for groups
    """
    return request.param


@pytest.fixture(scope='class', params=[-0.1, -0.5, -0.01])
def param_dist(request):
    """
    different values for distance parameters
    """
    return request.param


@pytest.fixture(scope='class', params=[1.2, 10])
def param_saving_weights(request):
    """
    different values for saving categories
    """
    return request.param


@pytest.fixture(scope='class', params=range(2))
def zone_h(request):
    """
    different values for home zone
    """
    return request.param


@pytest.fixture(scope='class')
def folder(request):
    """
    different values for home zone
    """
    folder = tempfile.gettempdir()
    # removing the temp folder does not work, because h5-files are still locked
    # by other processes
    # use %TEMP% folder instead
    return folder


@pytest.fixture(scope='class')
def result_folder(folder):
    """Temp folder to store the results"""
    result_folder = os.path.join(folder, 'results')
    os.makedirs(result_folder, exist_ok=True)
    return result_folder


@pytest.fixture()
def params_file(folder):
    """ The params-file"""
    fn = 'params'
    file_path = os.path.join(folder, '{}.h5'.format(fn))
    return file_path


@pytest.fixture()
def matrix_file(folder):
    """ The params-file"""
    fn = 'matrices'
    file_path = os.path.join(folder, '{}.h5'.format(fn))
    return file_path


@pytest.fixture()
def zones_file(folder):
    """ The params-file"""
    fn = 'zonal_data'
    file_path = os.path.join(folder, '{}.h5'.format(fn))
    return file_path


@pytest.fixture()
def result_file(folder):
    """ The params-file"""
    fn = 'results'
    file_path = os.path.join(folder, '{}.h5'.format(fn))
    return file_path


@pytest.fixture()
def balancing_file(folder):
    """ The params-file"""
    fn = 'balancing'
    file_path = os.path.join(folder, '{}.h5'.format(fn))
    return file_path


@pytest.fixture()
def wiver_files(params_file, matrix_file, zones_file,
                result_file, balancing_file):
    files = {'params': params_file,
             'matrices': matrix_file,
             'zonal_data': zones_file,
             'results': result_file,
             'balancing': balancing_file,}
    return files


@pytest.fixture(scope='function')
def wiver(request):
    """
    WIVER-instance
    """
    sector_name = np.array(['DL', 'IND'])

    wiver = WIVER(n_groups=2,
                  n_zones=5,
                  n_time_slices=3,
                  n_savings_categories=10,
                  n_modes=3,
                  n_sectors=len(sector_name))

    # mode of group
    wiver.mode_g = np.array([2, 2])
    # define centroids of zones
    x = np.array([2, 5, 1, 6, 10])
    y = np.array([5, 2, 3, 1, 7])
    # calc travel time between zones
    dx = x - x[:, np.newaxis]
    dy = y - y[:, np.newaxis]
    dist = np.sqrt(dx**2 + dy**2)
    wiver.km_ij = dist
    t = dist * 3 + 1

    # travel time is identical for all modes
    wiver.travel_time_mij[:] = t
    # define savings intervals
    # the first interval represents NINF (no connection)
    # for this interval the saving weight is 0, wich leads to 'no connection'
    wiver.savings_bins_s = np.concatenate(([np.NINF],
                                           np.linspace(.15, .85, 8),
                                           [1]))
    wiver.savings_weights_gs[0] = np.linspace(0, 2, wiver.n_savings_categories)
    wiver.savings_weights_gs[1] = np.linspace(0, 5, wiver.n_savings_categories)

    wiver.source_potential_gh = np.array([[10, 20, 10, 0, 10],
                                         [100, 20, 10, 0, 0]])
    wiver.sink_potential_gj = np.array([[0, 100, 10, 0, 100],
                                        [0, 20, 10, 0, 0]])
    wiver.tour_rates_g = np.array([2, 3])
    wiver.stops_per_tour_g = np.array([3, 2])

    wiver.time_series_starting_trips_gs = np.array([[5, 3, 1],
                                                    [6, 2, 0]])
    wiver.time_series_linking_trips_gs = np.array([[2, 3, 2],
                                                   [1, 6, 1]])
    wiver.time_series_ending_trips_gs = np.array([[0, 3, 6],
                                                  [1, 2, 8]])

    wiver.zone_no = np.arange(10, 60, 10)
    wiver.zone_name = np.array(['{}-Stadt'.format(i)
                                for i in 'ABCDE'])


    wiver.sector_short = sector_name
    wiver.sector_g = np.array([0, 1])

    wiver.mode_name = np.array(['Rad', 'Pkw', 'OV'])
    wiver.modes = np.array(['R', 'P', 'O'])
    wiver.groups = np.arange(2)
    wiver.group_names = np.array(['Gruppe {}'.format(i) for i in wiver.groups])

    return wiver


class Test01_WiverData:
    """Test the WiverData"""
    def test_01_test_definitions(self, wiver):
        """Test the WiverData creation"""
        wiver.define_datasets()
        print(wiver.params)
        print(wiver.zonal_data)
        print(wiver.matrices)
        print(wiver.results)
        # assert that the group names are set correctly
        np.testing.assert_array_equal(wiver.group_names, ['OV_DL', 'OV_IND'])

    def test_02_test_merge_definitions(self, wiver):
        """Test the WiverData creation"""
        wiver.define_datasets()
        wiver.merge_datasets()
        print(wiver.data)
        dims = wiver.data.dims
        assert dims['origins'] == wiver.n_zones
        assert dims['destinations'] == wiver.n_zones
        assert dims['modes'] == wiver.n_modes
        assert dims['groups'] == wiver.n_groups
        assert dims['savings'] == wiver.n_savings_categories
        assert dims['time_slices'] == wiver.n_time_slices

    def test_03_save_data(self, wiver, wiver_files):
        """Dave data as test data"""
        folder = os.path.dirname(wiver_files['params'])
        print('save to {}'.format(folder))
        wiver.define_datasets()
        wiver.calc()
        wiver.save_all_data(wiver_files)

    def test_04_read_data(self, wiver, wiver_files):
        """read model from test data and compare results with original data"""
        folder = os.path.dirname(wiver_files['params'])
        print('read from {}'.format(folder))
        wiver_from_file = WIVER.read_from_netcdf(wiver_files)
        wiver_from_file.calc()
        wiver.calc()
        np.testing.assert_almost_equal(wiver.trips_msij,
                                       wiver_from_file.trips_msij)

    def test_05_data_consistency(self, wiver):
        """Test the data consistency"""
        wiver.assert_data_consistency()
        # introduce unfeasable data to mode_g
        wiver.mode_g[0] = 3
        with pytest.raises(DataConsistencyError) as e:
            wiver.assert_data_consistency()
        print(e.value)

        wiver.mode_g[:] = (1, -1)
        with pytest.raises(DataConsistencyError) as e:
            wiver.assert_data_consistency()
        print(e.value)

    def test_06_save_data_as_visum(self, wiver, folder):
        """Dave results as Visum data"""
        print('save to {}'.format(folder))
        wiver.define_datasets()
        wiver.calc()
        wiver.save_results_to_visum(folder, visum_format='B')
        wiver.save_results_to_visum(folder, visum_format='V')
        wiver.save_results_to_visum(folder, visum_format='BK')


class Test02_Wiver:
    """Test Wiver Model"""

    def test_01_calc_tours(self, wiver, zone_h):
        """Test the tour rates and linking trips
        from different home zones (parameter: zone_h)"""
        g = 0
        tours = wiver.calc_tours(g, zone_h)
        target = wiver.source_potential_gh[g, zone_h] * wiver.tour_rates_g[g]
        assert tours == target
        linking_trips = wiver.calc_linking_trips(g, tours)
        assert linking_trips == tours * (wiver.stops_per_tour_g[g] - 1)
        print('{h}: {t}, {l}'.format(h=zone_h, t=tours, l=linking_trips))

    def test_02_destination_choice(self, wiver, group, param_dist):
        """Test the destination choice model"""
        h = 0
        t = 0
        wiver.calc_destination_choice(t, group, h)
        before = wiver.p_destination_tij[t, h].copy()
        print(before)

        wiver.param_dist_g[group] = param_dist
        wiver.calc_destination_choice(t, group, h)
        after = wiver.p_destination_tij[t, h].copy()
        print(after)

    def test_03_savings(self, wiver, zone_h):
        """Test the savings function for closeby points 1 and 3
        from different home zones (parameter: zone_h)"""
        g = 0
        i = 1
        j = 3
        m = wiver.mode_g[g]
        s = wiver.calc_savings(g, m, zone_h, i, j)
        t_ki = wiver.travel_time_mij[m, zone_h, i]
        t_jk = wiver.travel_time_mij[m, j, zone_h]
        t_ij = wiver.travel_time_mij[m, i, j]
        s_target = (t_ki + t_jk - t_ij) / (t_ki + t_jk)
        print(s)
        assert s == s_target

    def test_04_savings_factor(self, wiver, zone_h):
        """Test the savings factor function"""
        g = 0
        i = 1
        j = 3
        m = wiver.mode_g[g]
        sf = wiver.calc_savings_factor(g, m, zone_h, i, j)
        print(sf)

    def test_05_linking_trip_choice(self, wiver, group):
        """Test the destination choice model"""
        h = 0
        t = 0
        wiver.calc_destination_choice(t, group, h)
        wiver.calc_linking_trip_choice(t, group, h)
        print(wiver.p_links_tij[t])

    def test_06_calc_trips(self, wiver, group):
        """Test the destination choice model"""
        h = 0
        t = 0
        tours = 100
        linking_trips = 200
        wiver.calc_destination_choice(t, group, h)
        wiver.calc_linking_trip_choice(t, group, h)
        wiver.calc_trips(t, group, h, tours, linking_trips)
        np.testing.assert_almost_equal(wiver.home_based_trips_gij[group].sum(),
                                       tours)
        np.testing.assert_almost_equal(wiver.linking_trips_gij[group].sum(),
                                       linking_trips)

    def test_07_savings_influence(self, wiver):
        """Test the destination choice model"""

        def check_trips_with_savings(wiver, group, h, t, tours, linking_trips):
            wiver.home_based_trips_gij[group] = 0
            wiver.linking_trips_gij[group] = 0
            wiver.calc_linking_trip_choice(t, group, h)
            wiver.calc_trips(t, group, h, tours, linking_trips)
            link_matrix = wiver.linking_trips_gij[group]
            print(link_matrix.astype(int))
            np.testing.assert_almost_equal(link_matrix.sum(),
                                           linking_trips)
            print(link_matrix[0].sum())
            print(link_matrix[1].sum())
            zone_interior_link_trips = link_matrix.diagonal().sum()
            return zone_interior_link_trips

        h = 0
        t = 0
        group = 1
        tours = 100
        linking_trips = 200
        n_sav = wiver.n_savings_categories
        wiver.calc_destination_choice(t, group, h)

        # test first without any savings
        wiver.savings_weights_gs[group] = 1
        interior_trips_no_savings = check_trips_with_savings(
            wiver, group, h, t, tours, linking_trips)

        # test first with low savings
        wiver.savings_weights_gs[group] = np.linspace(1, 2, num=n_sav)
        interior_trips_low_savings = check_trips_with_savings(
            wiver, group, h, t, tours, linking_trips)

        # test first with medium savings
        wiver.savings_weights_gs[group] = np.linspace(1, 5, num=n_sav)
        interior_trips_medium_savings = check_trips_with_savings(
            wiver, group, h, t, tours, linking_trips)

        # test first with strong savings
        wiver.savings_weights_gs[group] = np.linspace(1, 10, num=n_sav)
        interior_trips_high_savings = check_trips_with_savings(
            wiver, group, h, t, tours, linking_trips)

        # make sure that the zonal interior trips should be more used
        # the higher the savings are

        print(interior_trips_no_savings, interior_trips_low_savings,
              interior_trips_medium_savings, interior_trips_high_savings)

        assert (interior_trips_no_savings < interior_trips_low_savings <
                interior_trips_medium_savings < interior_trips_high_savings)

    def test_08_calc_time_series(self, wiver):
        """Test the time_series"""
        wiver.calc_daily_trips()
        wiver.calc_time_series()

        # test the normalisation of the time series
        np.testing.assert_almost_equal(
            wiver.time_series_starting_trips_gs.sum(-1), 1)
        np.testing.assert_almost_equal(
            wiver.time_series_linking_trips_gs.sum(-1), 1)
        np.testing.assert_almost_equal(
            wiver.time_series_ending_trips_gs.sum(-1), 1)

        # test that the trips of the time slices add up to the daily matrix
        actual = wiver.trips_gsij.sum(1)
        desired = wiver.trips_gij
        np.testing.assert_array_almost_equal(actual, desired)

    def test_09_trips_gij(self, wiver):
        """Test if the arrays are not overwritten"""
        arr_gij_before = wiver.trips_gij.__array_interface__
        wiver.calc_daily_trips()
        arr_gij_after = wiver.trips_gij.__array_interface__
        assert arr_gij_before == arr_gij_after

    def test_10_wiver_calc(self, wiver):
        """Test the whole model"""
        wiver.calc()
        print(wiver.trips_gsij.sum(-1).sum(-1))

    def test_21_no_destinations(self, wiver):
        """Test exceptions when there is demand
        but no destinations accessible"""

        h = 0

        # normal execution without error
        t = 0
        group = 0
        wiver.calc_destination_choice(t, group, h)
        print(wiver.p_destination_tij[t, h])

        # now set sink_potential for group 0 to 0
        wiver.sink_potential_gj[0] = 0

        # this should raise a DestinationChoiceError, because no destinations
        # exists for group 0
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc_destination_choice(t, group, h)
        print(e.value)

        # make no destinations available
        t = 1
        group = 1
        m = wiver.mode_g[group]
        # infinitive travel times from home zone 0
        wiver.travel_time_mij[m, h, :] = np.inf
        # this should raise a Destination Choice Error, because there are
        # destinations for group 1, but they cannot be reached
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc_destination_choice(t, group, h)
        print(e.value)

    def test_22_no_accessible_linking_trips(self, wiver):
        """Test exceptions when there linking trips find no accessible destinations
        but no destinations accessible"""
        # make no destinations available
        h = 0
        t = 1
        group = 1
        m = wiver.mode_g[group]
        # infinitive travel times for linking trips
        wiver.travel_time_mij[m, 1:3, 1:3] = np.inf
        wiver.calc_destination_choice(t, group, h)
        # calculating the linking trips should raise an error
        # because of lack of accessibility between the destinations
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc_linking_trip_choice(t, group, h)
        print(e.value)

        # one accessible linking trip is suffitient
        wiver.travel_time_mij[m, 2, 1] = 10
        wiver.calc_destination_choice(t, group, h)
        wiver.calc_linking_trip_choice(t, group, h)
        assert wiver.p_links_tij[t, 2, 1] == 1

    def test_23_test_raises_destination_choice_error(self, wiver):
        """Test if the error is raised in the calc function"""
        wiver.calc()

        g = 0
        wiver.sink_potential_gj[g] = 0
        # this should raise a DestinationChoiceError, because no destinations
        # exists for group 0
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc()
        print(e.value)
        # reset sink_potential
        wiver.sink_potential_gj[g] = np.array([0, 100, 10, 0, 100])
        wiver.calc()

        g = 1
        h = 0
        m = wiver.mode_g[g]
        # infinitive travel times from home zone 0
        backup = wiver.travel_time_mij[m].copy()
        wiver.travel_time_mij[m, :, [1, 2]] = np.inf
        # this should raise a Destination Choice Error, because there are
        # destinations for group 1, but they cannot be reached
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc()
        print(e.value)
        # reset travel times
        wiver.travel_time_mij[m, :] = backup
        wiver.calc()

        # infinitive travel times for linking trips
        wiver.travel_time_mij[m, 1:3, 1:3] = np.inf
        # calculating the linking trips should raise an error
        # because of lack of accessibility between the destinations
        with pytest.raises(DestinationChoiceError) as e:
            wiver.calc()
        print(e.value)

    def test_31_aggregate_to_modes(self, wiver):
        """aggregate to modes"""
        wiver.calc()
        wiver.aggregate_to_modes()

        # test for daily matrix
        actual = wiver.trips_mij[2]
        # both groups are of mode 2
        desired = wiver.trips_gij.sum(0)
        np.testing.assert_array_almost_equal(actual, desired)

        # test for matrix by time-slices
        actual = wiver.trips_msij[2]
        desired = wiver.trips_gsij.sum(0)
        np.testing.assert_array_almost_equal(actual, desired)

        # modes without group should be 0
        np.testing.assert_array_equal(wiver.trips_mij[0:2], 0)
        np.testing.assert_array_equal(wiver.trips_msij[0:2], 0)

    def test_40_mean_distance_beta(self, wiver, param_dist):
        """
        Test the mean distance calculation
        with different distance parameters
        """
        wiver.param_dist_g[0] = param_dist
        wiver.calc()
        after = wiver.mean_distance_g
        print(after)

    def test_41_mean_distance_savings(self, wiver, param_saving_weights):
        """
        Test the mean distance calculation
        with different savings optimizations
        """
        wiver.savings_weights_gs[0] = np.linspace(1, param_saving_weights,
                                                  wiver.n_savings_categories)
        wiver.calc()
        after = wiver.mean_distance_g
        print(after)

    def test_42_mean_distance_modes(self, wiver, param_dist):
        """
        Test the mean distance modes calculation
        with different distance parameters
        """
        wiver.param_dist_g[0] = param_dist
        wiver.calc()
        after = wiver.mean_distance_m
        print(after)

    def test_50_balancing(self, wiver):
        """
        Test if the trip distribution equals the sink_potential
        and that the total number of trips stay constant
        """
        wiver.define_datasets()
        sp = wiver.zonal_data.sink_potential
        target_share = sp / sp.sum('destinations')
        wiver.calc_with_balancing(max_iterations=1)
        target_total_trips = wiver.trips_gij.sum()
        wiver.calc_with_balancing(max_iterations=20)
        trips = wiver.balancing.trips_to_destination
        actual_share = trips / trips.sum('destinations')
        np.testing.assert_allclose(actual_share, target_share, rtol=.05)

        actual_total_trips = wiver.trips_gij.sum()
        np.testing.assert_allclose(actual_total_trips,
                                   target_total_trips,
                                   rtol=.01)

    def test_51_test_starting_equals_ending_trips(self, wiver):
        """Test that starting and ending trips per zone are equal"""
        wiver.calc()
        starting_trips = wiver.trips_gij.sum(axis=2)
        ending_trips = wiver.trips_gij.sum(axis=1)
        np.testing.assert_almost_equal(starting_trips,
                                       ending_trips)

        # introduce an asymetry into the travel time matrix
        wiver.travel_time_mij[:, 0, :] *= 0.8
        wiver.travel_time_mij[:, 1, :] *= 1.2
        wiver.calc()
        starting_trips = wiver.trips_gij.sum(axis=2)
        ending_trips = wiver.trips_gij.sum(axis=1)
        np.testing.assert_almost_equal(starting_trips,
                                           ending_trips)


class Test03_TestExport:
    """Test export results"""
    def test_10_wiver_results(self, wiver, wiver_files):
        """Test export of the results"""
        wiver.calc()
        wiver.define_datasets()
        wiver.save_results(wiver_files)

    def test_11_wiver_detailed_results(self, wiver, result_folder):
        """Test export of the results to visum"""
        wiver.calc()
        wiver.define_datasets()
        wiver.save_detailed_results_to_visum(result_folder)

    def test_12_calc_starting_ending_trips(self, wiver):
        """Test calculation of starting and ending trips"""
        wiver.define_datasets()
        wiver.merge_datasets()
        wiver.calc_with_balancing()
        df = wiver.calc_starting_and_ending_trips()
        actual = df['start_end']
        target = df['modelled_trips']
        # test if with balancing the target trips and the actuals trips
        # do not differ more than 10 %
        np.testing.assert_allclose(actual, target, rtol=0.1)
        print(df)

    def test_20_run_wiver(self, folder):
        """Test wiver.run_wiver"""
        backup_sys_argv = sys.argv
        #folder = os.path.join(os.path.dirname(__file__), 'project')
        os.makedirs(folder, exist_ok=True)
        matrix_folder = os.path.join(folder, 'matrices')
        os.makedirs(matrix_folder, exist_ok=True)
        os.makedirs(os.path.join(folder, 'log'), exist_ok=True)

        sys.argv = ['', '-f={}'.format(folder), '-m={}'.format(matrix_folder)]
        try:
            gl = runpy.run_module('wiver.run_wiver', run_name='__main__')
        finally:
            sys.argv = backup_sys_argv



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import pytest

import xarray as xr
import numpy as np
import pandas as pd
import os
import orca
from wiver.wiver_python import WIVER
#from wiver.tests.filename_definitions import dataset_fn, zones_fn
from pytest_benchmark.plugin import benchmark


@pytest.fixture(scope='class', params=[5, 10, ])
def n_zones(request):
    return request.param


@pytest.fixture(scope='class', params=[1, 2, 4, 8, 16])
def n_groups(request):
    return request.param


class TestWiver:
    """Test Wiver Model"""

    def create_wiver(self, n_zones, n_groups):
        """create the wiver object"""
        wiver = WIVER(n_groups, n_zones, n_modes=1,
                      n_time_slices=2, n_savings_categories=9)
        # define centroids of zones
        x = np.random.randint(0, 50, (n_zones))
        y = np.random.randint(0, 50, (n_zones))
        # calc travel time between zones
        dx = x - x[:, np.newaxis]
        dy = y - y[:, np.newaxis]
        dist = np.sqrt(dx**2 + dy**2)
        t = dist * 3 + 1

        wiver.mode_g = np.zeros((n_groups, ))

        wiver.travel_time_mij[:] = t
        # define savings intervals
        wiver.savings_bins_s = np.concatenate((np.linspace(.15, .85, 8), [1]))
        for g in range(n_groups):
            wiver.savings_weights_gs[g] = np.linspace(
                1, (g + 2), wiver.n_savings_categories)

        wiver.source_potential_gh = np.random.randint(
            0, 10, (n_groups, n_zones))
        wiver.sink_potential_gj = np.random.randint(
            0, 10, (n_groups, n_zones))
        wiver.tour_rates_g = np.linspace(1, 3, num=n_groups)
        wiver.stops_per_tour_g = np.linspace(2, 4, num=n_groups)
        return wiver

    def test_01_calc_wiver(self, n_zones, n_groups):
        """Test the results of wiver for more zones"""
        wiver = self.create_wiver(n_zones, n_groups)
        # set zone 3 to 0
        wiver.source_potential_gh[:, 3] = 0
        wiver.sink_potential_gj[:, 3] = 0
        wiver.calc()
        # assure that the total number of trips per group fits
        actual = wiver.trips_gij.sum(-1).sum(-1)
        desired = (wiver.source_potential_gh.sum(-1) *
                   wiver.tour_rates_g * (wiver.stops_per_tour_g + 1))
        np.testing.assert_almost_equal(actual, desired)
        # assure that there are no trips from and to zone 3
        np.testing.assert_equal(wiver.trips_gij[:, 3], 0)
        np.testing.assert_equal(wiver.trips_gij[:, :, 3], 0)

        # assure that at least as many tours
        # from and to home zones start and end
        desired = wiver.source_potential_gh * wiver.tour_rates_g[:, np.newaxis]
        actual = wiver.trips_gij.sum(1)
        np.testing.assert_array_equal((actual - desired) > -0.0000001, 1)

        actual = wiver.trips_gij.sum(2)
        np.testing.assert_array_equal((actual - desired) > -0.0000001, 1)

    def test_02_benchmark_wiver(self, benchmark, n_zones, n_groups):
        """Test the wiver runtime for more zones"""
        wiver = self.create_wiver(n_zones, n_groups)
        benchmark(wiver.calc)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 20:33:08 2016

@author: MaxBohnet
"""

from setuptools import setup, find_packages
from cythoninstallhelpers.make_cython_extensions import make_extensions
from cythoninstallhelpers.get_version import get_version


package_name = 'wiver'
version = get_version(package_name, __file__)
ext_modnames = ['wiver.wiver_cython',
                ]


setup(
    name=package_name,
    version=version,
    description="commercial trip model and pessenger demand model",

    packages=find_packages('src', exclude=['ez_setup']),
    #namespace_packages=['wiver'],

    package_dir={'': 'src'},
    package_data={'': ['*.pxd']},
    include_package_data=True,
    zip_safe=False,
    data_files=[
        ],

    ext_modules=make_extensions(ext_modnames),
)
